cdef extern from "interp_common.c":
    Py_ssize_t c_binary_search_with_guess(
            double key,
            double* array,
            Py_ssize_t length,
            Py_ssize_t guess
            ) noexcept nogil

cdef double interp_ptr(
        double desired_x,
        double* x_domain,
        double* dependent_values,
        Py_ssize_t len_x,
        Py_ssize_t provided_j = *
        ) noexcept nogil

cdef double complex interp_complex_ptr(
        double desired_x,
        double* x_domain,
        double complex* dependent_values,
        Py_ssize_t len_x,
        Py_ssize_t provided_j = *
        ) noexcept nogil

cpdef double interp(
        double desired_x,
        double[:] x_domain,
        double[:] dependent_values,
        Py_ssize_t provided_j = *
        ) noexcept nogil

cpdef double complex interp_complex(
        double desired_x,
        double[:] x_domain,
        double complex[:] dependent_values,
        Py_ssize_t provided_j = *
        ) noexcept nogil

cdef (double, Py_ssize_t) interpj_ptr(
        double desired_x,
        double* x_domain,
        double* dependent_values,
        Py_ssize_t len_x,
        Py_ssize_t provided_j = *
        ) noexcept nogil

cpdef (double, Py_ssize_t) interpj(
        double desired_x,
        double[:] x_domain,
        double[:] dependent_values,
        Py_ssize_t provided_j = *
        ) noexcept nogil

cdef (double complex, Py_ssize_t) interp_complexj_ptr(
        double desired_x,
        double* x_domain,
        double complex* dependent_values,
        Py_ssize_t len_x,
        Py_ssize_t provided_j = *
        ) noexcept nogil

cpdef (double complex, Py_ssize_t) interp_complexj(
        double desired_x,
        double[:] x_domain,
        double complex[:] dependent_values,
        Py_ssize_t provided_j = *
        ) noexcept nogil

cdef void interp_array_ptr(
        double* desired_x_array,
        double* x_domain,
        double* dependent_values,
        double* desired_dependent_array,
        Py_ssize_t len_x,
        Py_ssize_t desired_len
        ) noexcept nogil

cpdef void interp_array(
        double[::1] desired_x_array,
        double[::1] x_domain,
        double[::1] dependent_values,
        double[::1] desired_dependent_array
        ) noexcept nogil

cdef void interp_complex_array_ptr(
        double* desired_x_array,
        double* x_domain,
        double complex* dependent_values,
        double complex* desired_dependent_array,
        Py_ssize_t len_x,
        Py_ssize_t desired_len
        ) noexcept nogil

cpdef void interp_complex_array(
        double[::1] desired_x_array,
        double[::1] x_domain,
        double complex[::1] dependent_values,
        double complex[::1] desired_dependent_array
        ) noexcept nogil


from CyRK.cy.cysolver_api cimport cysolve_ivp, cysolve_ivp_gil, cysolve_ivp_noreturn, DiffeqFuncType, PreEvalFunc, CySolverResult, WrapCySolverResult, CySolverBase, CySolveOutput, RK23_METHOD_INT, RK45_METHOD_INT, DOP853_METHOD_INT, MAX_STEP
from CyRK.cy.helpers cimport interpolate_from_solution_list


import numpy as np
from numba import njit

initial_conds = np.asarray((20., 20.), dtype=np.complex128)
initial_conds_float = np.asarray((20., 20.), dtype=np.float64)
time_span = (0., 20.)
rtol = 1.0e-7
atol = 1.0e-8


@njit
def cy_diffeq(t, y, dy):
    dy[0] = (1. - 0.01 * y[1]) * y[0]
    dy[1] = (0.02 * y[0] - 1.) * y[1]

@njit
def pysolve_diffeq(dy, t, y):
    dy[0] = (1. - 0.01 * y[1]) * y[0]
    dy[1] = (0.02 * y[0] - 1.) * y[1]

@njit
def nb_diffeq(t, y):
    dy = np.empty_like(y)
    dy[0] = (1. - 0.01 * y[1]) * y[0]
    dy[1] = (0.02 * y[0] - 1.) * y[1]
    return dy


def test_nbrk():

    from CyRK import nbsolve_ivp

    nbrk_result = nbsolve_ivp(nb_diffeq, time_span, initial_conds)

    assert nbrk_result.success
    assert type(nbrk_result.t) == np.ndarray
    assert type(nbrk_result.y) == np.ndarray
    assert nbrk_result.y.shape[0] == 2

    print("CyRK's nbsolve_ivp was tested successfully.")

def test_pysolver():

    from CyRK import pysolve_ivp

    result = pysolve_ivp(pysolve_diffeq, time_span, initial_conds_float, pass_dy_as_arg=True)

    assert result.success
    assert type(result.t) == np.ndarray
    assert type(result.y) == np.ndarray
    assert result.y.shape[0] == 2

    print("CyRK's PySolver was tested successfully.")

def test_cysolver():

    from CyRK.cy.cysolver_test import cytester

    result = cytester(0,
                      time_span,
                      initial_conds_float,
                      args=None,
                      method=1,
                      expected_size=0,
                      max_num_steps=0,
                      max_ram_MB=2000,
                      rtol=rtol,
                      atol=atol)

    assert result.success
    assert type(result.t) == np.ndarray
    assert type(result.y) == np.ndarray
    assert result.y.shape[0] == 2

    print("CyRK's CySolver was tested successfully.")



from libcpp cimport bool as cpp_bool
cimport cpython.ref as cpy_ref

from CyRK.utils.memory cimport shared_ptr
from CyRK.cy.pysolver_cyhook cimport DiffeqMethod
from CyRK.cy.cysolver_api cimport CySolverResult, NowStatePointers

cimport numpy as cnp
cnp.import_array()

# =====================================================================================================================
# Import the C++ cysolve_ivp helper function
# =====================================================================================================================
cdef extern from "cysolve.cpp" nogil:
    # Python-hook implementation

    cdef cppclass PySolver:
        PySolver()
        PySolver(
            int integration_method,
            cpy_ref.PyObject* cython_extension_class_instance,
            DiffeqMethod cython_extension_class_diffeq_method,
            shared_ptr[CySolverResult] solution_sptr,
            const double t_start,
            const double t_end,
            const double* y0_ptr,
            const size_t num_y,
            const size_t expected_size,
            const size_t num_extra,
            const size_t max_num_steps,
            const size_t max_ram_MB,
            const cpp_bool dense_output,
            const double* t_eval,
            const size_t len_t_eval,
            const double rtol,
            const double atol,
            const double* rtols_ptr,
            const double* atols_ptr,
            const double max_step_size,
            const double first_step_size)
        void solve()
        int status
        int integration_method
        shared_ptr[CySolverResult] solution_sptr


cdef class WrapPyDiffeq:

    cdef object diffeq_func
    cdef tuple args
    cdef cpp_bool use_args
    cdef cpp_bool pass_dy_as_arg

    cdef size_t num_y
    cdef size_t num_dy

    cdef cnp.ndarray y_now_arr
    cdef double[::1] y_now_view
    cdef cnp.ndarray dy_now_arr
    cdef double[::1] dy_now_view

    # State attributes
    cdef double* y_now_ptr
    cdef double* t_now_ptr
    cdef double* dy_now_ptr

    cdef void set_state(self, NowStatePointers* solver_state_ptr) noexcept
    
    cdef void diffeq(self) noexcept



# distutils: language = c++
# cython: boundscheck=False, wraparound=False, nonecheck=False, cdivision=True, initializedcheck=False

from libc.string cimport memcpy
from libcpp.cmath cimport fmin, fabs

from CyRK.utils.memory cimport make_shared
from CyRK.cy.cysolver_api cimport find_expected_size, WrapCySolverResult, INF, EPS_100

import numpy as np
cimport numpy as cnp
cnp.import_array()

# =====================================================================================================================
# PySolver Class (holds the intergrator class and reference to the python diffeq function)
# =====================================================================================================================
cdef class WrapPyDiffeq:

    def __cinit__(
            self,
            object diffeq_func,
            tuple args,
            size_t num_y,
            size_t num_dy,
            bint pass_dy_as_arg = False
            ):
        
        # Install differential equation function and any additional args
        self.diffeq_func = diffeq_func
        if args is None:
            self.args     = None
            self.use_args = False
        else:
            if len(args) == 0:
                # Empty tuple provided. Don't use args.
                self.args     = None
                self.use_args = False
            else:
                self.args     = args
                self.use_args = True
        
        # Build python-safe arrays
        self.num_y  = num_y
        self.num_dy = num_dy

        if pass_dy_as_arg:
            self.pass_dy_as_arg = True
        else:
            self.pass_dy_as_arg = False
    
    cdef void set_state(self, NowStatePointers* solver_state_ptr) noexcept:
        
        self.t_now_ptr  = solver_state_ptr.t_now_ptr
        self.y_now_ptr  = solver_state_ptr.y_now_ptr
        self.dy_now_ptr = solver_state_ptr.dy_now_ptr

        # Create memoryviews of the pointers
        self.y_now_view  = <double[:self.num_y]>self.y_now_ptr

        # Create numpy arrays which will be passed to the python diffeq.
        # We need to make sure that this is a not a new ndarray, but one that points to the same data. 
        # That is why we use `PyArray_SimpleNewFromData` instead of a more simple `asarray`.
        # Note that it is not safe to return these arrays outside of this class because they may get deallocated while
        # the numpy array still points to the underlying memory.
        cdef cnp.npy_intp[1] shape
        cdef cnp.npy_intp* shape_ptr = &shape[0]
        shape_ptr[0] = <cnp.npy_intp>self.num_y
        
        self.y_now_arr = cnp.PyArray_SimpleNewFromData(1, shape_ptr, cnp.NPY_DOUBLE, self.y_now_ptr)
        
        # Do the same for dy if the user provided the appropriate kind of differential equation.
        if self.pass_dy_as_arg:
            self.dy_now_view = <double[:self.num_dy]>self.dy_now_ptr
            shape[0]         = <cnp.npy_intp>self.num_dy  # dy may have a larger shape than y
            self.dy_now_arr  = cnp.PyArray_SimpleNewFromData(1, shape_ptr, cnp.NPY_DOUBLE, self.dy_now_ptr)   

    cdef void diffeq(self) noexcept:
        # Run python diffeq
        if self.pass_dy_as_arg:
            if self.use_args:
                self.diffeq_func(self.dy_now_arr, self.t_now_ptr[0], self.y_now_arr, *self.args)
            else:
                self.diffeq_func(self.dy_now_arr, self.t_now_ptr[0], self.y_now_arr)
        else:
            if self.use_args:
                self.dy_now_view = self.diffeq_func(self.t_now_ptr[0], self.y_now_arr, *self.args)
            else:
                self.dy_now_view = self.diffeq_func(self.t_now_ptr[0], self.y_now_arr)
            # Since we do not have a static dy that we can pass to the function and use in the solver we must copy over
            # the values from the newly created dy memory view
            # Note that num_dy may be larger than num_y if the user is capturing extra output during integration.
            memcpy(self.dy_now_ptr, &self.dy_now_view[0], sizeof(double) * self.num_dy)

# =====================================================================================================================
# PySolver wrapper function
# =====================================================================================================================

def pysolve_ivp(
        object py_diffeq,
        tuple time_span,
        const double[::1] y0,
        str method = 'RK45',
        const double[::1] t_eval = None,
        bint dense_output = False,
        tuple args = None,
        size_t expected_size = 0,
        size_t num_extra = 0,
        double first_step = 0.0,
        double max_step = INF,
        rtol = 1.0e-3,
        atol = 1.0e-6,
        size_t max_num_steps = 0,
        size_t max_ram_MB = 2000,
        bint pass_dy_as_arg = False
        ):

    # Parse method
    method = method.lower()
    cdef int integration_method = 1
    if method == "rk23":
        integration_method = 0
    elif method == "rk45":
        integration_method = 1
    elif method == 'dop853':
        integration_method = 2
    else:
        raise NotImplementedError(
            f"Unknown or unsupported integration method provided: {method}.\n"
            f"Supported methods are: RK23, RK45, DOP853."
            )
    
    # Parse time_span
    cdef double t_start = time_span[0]
    cdef double t_end   = time_span[1]
    cdef cpp_bool direction_flag = (t_end - t_start) >= 0

    # Parse y0
    cdef size_t num_y   = len(y0)
    cdef const double* y0_ptr = &y0[0]
    
    # Parse t_eval
    cdef const double* t_eval_ptr = NULL
    cdef cpp_bool t_eval_provided = False
    cdef size_t len_t_eval = 0
    if t_eval is not None:
        t_eval_provided = True
        len_t_eval = len(t_eval)
        t_eval_ptr = &t_eval[0]
    
    # Parse num_extra
    cdef size_t num_dy = num_y + num_extra
    
    # Parse rtol
    cdef double* rtols_ptr = NULL
    cdef double[::1] rtols_view
    cdef double rtol_float = 0.0
    if type(rtol) == float:
        rtol_float = rtol
        rtol_float = fabs(rtol_float)
    else:
        rtols_view = np.abs(np.asarray(rtol, dtype=np.float64, order='C'))
        rtols_ptr = &rtols_view[0]
    
    # Parse atol
    cdef double* atols_ptr = NULL
    cdef double[::1] atols_view
    cdef double atol_float = 0.0
    if type(atol) == float:
        atol_float = atol
        atol_float = fabs(atol_float)
    else:
        atols_view = np.abs(np.asarray(atol, dtype=np.float64, order='C'))
        atols_ptr = &atols_view[0]
    
    # Parse expected size
    cdef size_t expected_size_touse = expected_size
    cdef double rtol_tmp
    cdef double min_rtol = INF
    if expected_size_touse == 0:
        if rtols_ptr:
            # rtol for each y
            for y_i in range(num_y):
                rtol_tmp = rtols_ptr[y_i]
                if rtol_tmp < EPS_100:
                    rtol_tmp = EPS_100
                min_rtol = fmin(min_rtol, rtol_tmp)
        else:
            # Only one rtol
            rtol_tmp = rtol
            if rtol_tmp < EPS_100:
                rtol_tmp = EPS_100
            min_rtol = rtol_tmp
        expected_size_touse = find_expected_size(num_y, num_extra, fabs(t_end - t_start), min_rtol)
    
    # Parse first step size
    if first_step < 0.0:
        raise AttributeError("First step size must be a postive float (or 0.0 to use automatic finder).")

    # Parse maximum step size
    if max_step <= 0.0:
        raise AttributeError("Maximum step size must be a postive float.")

    # Build solution storage
    cdef shared_ptr[CySolverResult] solution_sptr = make_shared[CySolverResult](
        num_y,
        num_extra,
        expected_size,
        t_end,
        direction_flag,
        dense_output,
        t_eval_provided)

    # Build diffeq wrapper
    cdef WrapPyDiffeq diffeq_wrap = WrapPyDiffeq(py_diffeq, args, num_y, num_dy, pass_dy_as_arg=pass_dy_as_arg)
    cdef DiffeqMethod diffeq_func = <DiffeqMethod>diffeq_wrap.diffeq

    # Finally we can actually run the integrator!
    # The following effectively copies the functionality of cysolve_ivp. We can not directly use that function
    # because we need to tie in the python-based diffeq function (via its wrapper)

    # We need to heap allocate the PySolver class instance otherwise it can get garbage collected while the solver
    # is running.
    cdef PySolver* solver = new PySolver(
            integration_method,
            <cpy_ref.PyObject*>diffeq_wrap,
            diffeq_func,
            solution_sptr,
            t_start,
            t_end,
            y0_ptr,
            num_y,
            expected_size,
            num_extra,
            max_num_steps,
            max_ram_MB,
            dense_output,
            t_eval_ptr,
            len_t_eval,
            rtol_float,
            atol_float,
            rtols_ptr,
            atols_ptr,
            max_step,
            first_step
        )

    # Get pointers to the solver so that the Python differential equation can use and update its state variables (t_now, y_now, dy_now).
    cdef NowStatePointers solver_state = solution_sptr.get().solver_uptr.get().get_now_state()
    diffeq_wrap.set_state(&solver_state)
    
    ##
    # Run the integrator!
    ##
    solver.solve()
    
    # Wrap the solution in a python-safe wrapper class
    cdef WrapCySolverResult pyresult = WrapCySolverResult()
    pyresult.set_cyresult_pointer(solution_sptr)

    del solver

    return pyresult



from typing import Tuple

import numpy as np
from numba import njit

from CyRK.nb.dop_coefficients import (
    A as A_DOP, B as B_DOP, C as C_DOP, E3 as E3_DOP, E5 as E5_DOP, D as D_DOP,
    N_STAGES as N_STAGES_DOP, N_STAGES_EXTENDED as N_STAGES_EXTENDED_DOP, ORDER as ORDER_DOP,
    ERROR_ESTIMATOR_ORDER as ERROR_ESTIMATOR_ORDER_DOP)

from CyRK.nb.nb_storage import WrapNBRKResult

# Optimizations
EMPTY_ARR = np.empty(0, dtype=np.float64)
EPS = np.finfo(np.float64).eps
EPS_10 = 10. * EPS
EPS_100 = 100. * EPS
INF = np.inf

# Diffeq Solver Settings
# Multiply steps computed from asymptotic behaviour of errors by this.
SAFETY = 0.9

MIN_FACTOR = 0.2  # Minimum allowed decrease in a step size.
MAX_FACTOR = 10.  # Maximum allowed increase in a step size.

RK23_order = 3
RK23_error_estimator_order = 2
RK23_n_stages = 3
RK23_C = np.array([0, 1 / 2, 3 / 4], order='C', dtype=np.float64)
RK23_A = np.array(
        [
            [0, 0, 0],
            [1 / 2, 0, 0],
            [0, 3 / 4, 0]
            ], order='C', dtype=np.float64
        )
RK23_B = np.array([2 / 9, 1 / 3, 4 / 9], order='C', dtype=np.float64)
RK23_E = np.array([5 / 72, -1 / 12, -1 / 9, 1 / 8], order='C', dtype=np.float64)
RK23_P = np.array(
        [[1, -4 / 3, 5 / 9],
         [0, 1, -2 / 3],
         [0, 4 / 3, -8 / 9],
         [0, -1, 1]], order='C', dtype=np.float64
        )

RK45_order = 5
RK45_error_estimator_order = 4
RK45_n_stages = 6
RK45_C = np.array([0, 1 / 5, 3 / 10, 4 / 5, 8 / 9, 1], order='C', dtype=np.float64)
RK45_A = np.array(
        [
            [0, 0, 0, 0, 0],
            [1 / 5, 0, 0, 0, 0],
            [3 / 40, 9 / 40, 0, 0, 0],
            [44 / 45, -56 / 15, 32 / 9, 0, 0],
            [19372 / 6561, -25360 / 2187, 64448 / 6561, -212 / 729, 0],
            [9017 / 3168, -355 / 33, 46732 / 5247, 49 / 176, -5103 / 18656]
            ], order='C', dtype=np.float64
        )
RK45_B = np.array([35 / 384, 0, 500 / 1113, 125 / 192, -2187 / 6784, 11 / 84], order='C', dtype=np.float64)
RK45_E = np.array(
        [-71 / 57600, 0, 71 / 16695, -71 / 1920, 17253 / 339200, -22 / 525,
         1 / 40], order='C', dtype=np.float64
        )

RK45_P = np.array(
        [
            [1, -8048581381 / 2820520608, 8663915743 / 2820520608,
             -12715105075 / 11282082432],
            [0, 0, 0, 0],
            [0, 131558114200 / 32700410799, -68118460800 / 10900136933,
             87487479700 / 32700410799],
            [0, -1754552775 / 470086768, 14199869525 / 1410260304,
             -10690763975 / 1880347072],
            [0, 127303824393 / 49829197408, -318862633887 / 49829197408,
             701980252875 / 199316789632],
            [0, -282668133 / 205662961, 2019193451 / 616988883, -1453857185 / 822651844],
            [0, 40617522 / 29380423, -110615467 / 29380423, 69997945 / 29380423]], order='C', dtype=np.float64
        )


@njit(cache=False)
def _norm(x):
    return np.linalg.norm(x) / np.sqrt(x.size)


@njit(cache=False, fastmath=False)
def nbsolve_ivp(
        diffeq: callable,
        t_span: Tuple[float, float],
        y0: np.ndarray,
        args: tuple = tuple(),
        rtol: float = 1.e-3,
        atol: float = 1.e-6,
        rtols: np.ndarray = EMPTY_ARR,
        atols: np.ndarray = EMPTY_ARR,
        max_step: float = np.inf,
        first_step: float = None,
        rk_method: int = 1,
        t_eval: np.ndarray = EMPTY_ARR,
        capture_extra: bool = False,
        interpolate_extra: bool = False,
        max_num_steps: int = 0
        ):
    """ A Numba-safe Runge-Kutta Integrator based on Scipy's solve_ivp RK integrator.

    Parameters
    ----------
    diffeq : callable
        An njit-compiled function that defines the derivatives of the problem.
    t_span : Tuple[float, float]
        A tuple of the beginning and end of the integration domain's dependent variables.
    y0 : np.ndarray
        1D array of the initial values of the problem at t_span[0]
    args : tuple = tuple()
        Any additional arguments that are passed to dffeq.
    rtol : float = 1.e-3
        Integration relative tolerance used to determine optimal step size.
    atol : float = 1.e-6
        Integration absolute tolerance used to determine optimal step size.
    max_step : float = np.inf
        Maximum allowed step size.
    first_step : float = None
        Initial step size. If `None`, then the function will attempt to determine an appropriate initial step.
    rk_method : int = 1
        The type of RK method used for integration
            0 = RK23
            1 = RK45
            2 = DOP853
    t_eval : np.ndarray = None
        If provided, then the function will interpolate the integration results to provide them at the
            requested t-steps.
    capture_extra : bool = False
        If True, then extra output will be captured from the differential equation.
        See CyRK's Documentation/Extra Output.md for more information
    interpolate_extra : bool = False
        If True, then extra output will be interpolated (along with y) at t_eval. Otherwise, y will be interpolated
         and then differential equation will be called to find the output at each t in t_eval.
    max_num_steps : int = 0
        Maximum number of steps integrator is allowed to take.
        If set to 0 (the default) then an infinite number of steps are allowed.

    References
    ----------
    .. [1] E. Hairer, S. P. Norsett G. Wanner, "Solving Ordinary Differential
           Equations I: Nonstiff Problems", Sec. II.
    .. [2] `Page with original Fortran code of DOP853
            <http://www.unige.ch/~hairer/software.html>`_.

    Returns
    -------
    time_domain : np.ndarray
        The final time domain. This is equal to t_eval if it was provided.
    y_results : np.ndarray
        The solution of the differential equation provided for each time_result.
            If `capture_extra` was set to True then this will output both y and any extra parameters calculated by the
             differential equation. The format of this output will look like:
            ```
            y_results[0:y_size, :]          = ... # Actual y-results calculated by the diffeq solver
            y_results[y_size:extra_size, :] = ... # Extra outputs captured alongside y during integration

    success : bool
        Final integration success flag.
    message : str
        Any integration messages, useful if success=False.

    """

    # Clean up and interpret inputs
    t_start = t_span[0]
    t_end = t_span[1]
    direction = np.sign(t_end - t_start) if t_end != t_start else 1
    direction_inf = direction * np.inf
    y0 = np.asarray(y0)
    y_size = y0.size
    y_size_sqrt = np.sqrt(y_size)
    dtype = y0.dtype
    t_eval_size = t_eval.size
    run_interpolation = t_eval_size > 0
    store_extras_during_integration = capture_extra
    if run_interpolation and not interpolate_extra:
        # If y is eventually interpolated but the extra outputs are not being interpolated, then there is
        #  no point in storing the values during the integration. Turn off this functionality to save
        #  on computation
        store_extras_during_integration = False

    # If extra output is true then the output of the diffeq will be larger than the size of y0.
    #   determine that extra size by calling the diffeq and checking its size.
    extra_start = y_size
    extra_size = 0
    total_size = y_size
    if capture_extra:
        output_ = np.asarray(diffeq(t_start, y0, *args), dtype=dtype)
        total_size = output_.size
        extra_size = total_size - y_size
        extra_start = y_size

        # Extract the extra output from the function output.
        y0_plus_extra = np.empty(total_size, dtype=dtype)
        for i in range(total_size):
            if i < extra_start:
                # Pull from y0
                y0_plus_extra[i] = y0[i]
            else:
                # Pull from extra output
                y0_plus_extra[i] = output_[i]
        if store_extras_during_integration:
            y0_to_store = y0_plus_extra
            store_loop_size = total_size
        else:
            y0_to_store = y0
            store_loop_size = y_size

    else:
        y0_to_store = y0
        store_loop_size = y_size
    extra_result = np.empty(extra_size, dtype=dtype)

    if store_extras_during_integration:
        y_result_store = np.empty(total_size, dtype=dtype)
    else:
        y_result_store = np.empty(y_size, dtype=dtype)

    # Containers to store results during integration
    time_domain_list = [t_start]
    y_result_list = [np.copy(y0_to_store)]

    # Integrator Status Codes
    status = 0
    message = "Integration is/was ongoing (perhaps it was interrupted?)."

    # Determine RK constants
    if rk_method == 0:
        # RK23 Method
        rk_order = RK23_order
        error_order = RK23_error_estimator_order
        rk_n_stages = RK23_n_stages
        rk_n_stages_plus1 = rk_n_stages + 1
        C = RK23_C
        A = RK23_A
        B = RK23_B
        E = np.asarray(RK23_E, dtype=dtype)
        # TODO: Used in dense output calculation. Not needed until that is implemented.
        # P = RK23_P
        # Set these unused variables to E to avoid variable not set check
        E3 = E
        E5 = E

        # Initialize RK-K variable
        K = np.empty((rk_n_stages_plus1, y_size), dtype=dtype)
    elif rk_method == 1:
        # RK45 Method
        rk_order = RK45_order
        error_order = RK45_error_estimator_order
        rk_n_stages = RK45_n_stages
        rk_n_stages_plus1 = rk_n_stages + 1
        C = RK45_C
        A = RK45_A
        B = RK45_B
        E = np.asarray(RK45_E, dtype=dtype)
        # TODO: Used in dense output calculation. Not needed until that is implemented.
        # P = RK45_P
        # Set these unused variables to E to avoid variable not set check
        E3 = E
        E5 = E

        # Initialize RK-K variable
        K = np.empty((rk_n_stages_plus1, y_size), dtype=dtype)
    elif rk_method == 2:
        # DOP853
        rk_order = ORDER_DOP
        error_order = ERROR_ESTIMATOR_ORDER_DOP
        rk_n_stages = N_STAGES_DOP
        rk_n_stages_plus1 = rk_n_stages + 1
        A = A_DOP[:rk_n_stages, :rk_n_stages]
        B = B_DOP
        C = C_DOP[:rk_n_stages]
        E3 = np.asarray(E3_DOP, dtype=dtype)
        E5 = np.asarray(E5_DOP, dtype=dtype)
        # TODO: Used in dense output calculation. Not needed until that is implemented.
        # D = np.asarray(D_DOP, dtype=dtype)
        # A_EXTRA = np.asarray(A_DOP[rk_n_stages + 1:], dtype=dtype)
        # C_EXTRA = np.asarray(C_DOP[rk_n_stages + 1:], dtype=dtype)
        # Set these unused variables to E to avoid variable not set check
        E = E3

        # Initialize RK-K variable
        K_extended = np.empty((N_STAGES_EXTENDED_DOP, y_size), dtype=dtype)
        K = np.ascontiguousarray(K_extended[:rk_n_stages_plus1, :])
    else:
        status = -8
        message = "Attribute error."
        raise AttributeError(
            'Unexpected rk_method provided. Currently supported versions are:\n'
            '\t0 = RK23\n'
            '\t1 = RK34\n'
            '\t2 = DOP853')

    # Recast some constants into the correct dtype, so they can be used with y0.
    A = np.asarray(A, dtype=dtype)
    B = np.asarray(B, dtype=dtype)

    # Determine some RK-related constants
    len_C = len(C)
    A_at_10 = A[1, 0]
    error_expo = 1. / (error_order + 1.)

    # Setup tolerances
    rtol_array = np.empty(y_size, dtype=np.float64)
    rtol_size = rtols.size
    if rtol_size > 0:
        if rtol_size != y_size:
            raise AttributeError('rtols must be the same size as y0.')
        for i in range(y_size):
            rtol_ = rtols[i]
            if rtol_ < (100 * EPS):
                rtol_ = (100 * EPS)
            rtol_array[i] = rtol_
    else:
        if rtol < (100 * EPS):
            rtol = (100 * EPS)
        for i in range(y_size):
            rtol_array[i] = rtol

    atol_array = np.empty(y_size, dtype=np.float64)
    atol_size = atols.size
    if atol_size > 0:
        if atol_size != y_size:
            raise AttributeError('atols must be the same size as y0.')
        for i in range(y_size):
            atol_array[i] = atols[i]
    else:
        for i in range(y_size):
            atol_array[i] = atol

    # Determine maximum number of steps
    if max_num_steps == 0:
        use_max_steps = False
    elif max_num_steps < 0:
        raise AttributeError('Negative number of max steps provided.')
    else:
        use_max_steps = True

    # Initialize variables for start of integration
    t_old    = t_start
    t_new    = t_start
    y_new    = np.empty_like(y0)
    y_old    = np.empty_like(y0)
    dydt_old = np.empty_like(y0)
    dydt_new = np.empty_like(y0)
    for i in range(y_size):
        y0_i     = y0[i]
        y_new[i] = y0_i
        y_old[i] = y0_i

    output = np.asarray(diffeq(t_new, y_new, *args), dtype=dtype)
    for i in range(y_size):
        dydt_old[i] = output[i]

    # Find first step size
    first_step_found = False
    if first_step is not None:
        step_size = max_step
        if first_step < 0.:
            status = -8
            message = "Attribute error."
            raise AttributeError('Error in user-provided step size: Step size must be a positive number.')
        elif first_step > np.abs(t_end - t_start):
            status = -8
            message = "Attribute error."
            raise AttributeError('Error in user-provided step size: Step size can not exceed bounds.')
        elif first_step != 0.:
            step_size = first_step
            first_step_found = True

    if not first_step_found:
        # Select an initial step size based on the differential equation.
        # .. [1] E. Hairer, S. P. Norsett G. Wanner, "Solving Ordinary Differential
        #        Equations I: Nonstiff Problems", Sec. II.4.
        if y_size == 0:
            step_size = INF
        else:

            # Take the norm of d0 and d1
            d0 = 0.
            d1 = 0.
            for i in range(y_size):
                scale = atol_array[i] + np.abs(y_old[i]) * rtol_array[i]

                d0_abs = np.abs(y_old[i] / scale)
                d1_abs = np.abs(dydt_old[i] / scale)
                d0 += (d0_abs * d0_abs)
                d1 += (d1_abs * d1_abs)

            d0 = np.sqrt(d0) / y_size_sqrt
            d1 = np.sqrt(d1) / y_size_sqrt

            if d0 < 1.e-5 or d1 < 1.e-5:
                h0 = 1.e-6
            else:
                h0 = 0.01 * d0 / d1

            y1 = y_old + h0 * direction * dydt_old
            t1 = t_old + h0 * direction

            # Use the differential equation to estimate the first step size
            diffeq_output = np.asarray(diffeq(t1, y1, *args), dtype=dtype)
            for i in range(y_size):
                dydt_new[i] = diffeq_output[i]

            d2 = 0.
            for i in range(y_size):
                scale = atol_array[i] + np.abs(y_old[i]) * rtol_array[i]
                d2_abs = np.abs((dydt_new[i] - dydt_old[i]) / scale)
                d2 += (d2_abs * d2_abs)
            d2 = np.sqrt(d2) / (h0 * y_size_sqrt)

            if d1 <= 1.e-15 and d2 <= 1.e-15:
                h1 = max(1.e-6, h0 * 1.e-3)
            else:
                h1 = (0.01 / max(d1, d2))**error_expo

            next_after = 10. * abs(np.nextafter(t_old, direction * np.inf) - t_old)
            step_size  = max(next_after, min(100. * h0, h1))

    # Main integration loop
    if y_size == 0:
        status = -6
        message = "Integration never started: y-size is zero."

    # # Time Loop
    len_t = 1  # Already one time step due to initial conditions.
    while status == 0:

        if t_new == t_end:
            t_old = t_end
            t_new = t_end
            status = 1
            break

        if use_max_steps:
            if len_t > max_num_steps:
                status = -7
                message = "Maximum number of steps (set by user) exceeded during integration."
                break

        # Run RK integration step
        # Determine step size based on previous loop
        # Find minimum step size based on the value of t (less floating point numbers between numbers when t is large)
        next_after = 10. * abs(np.nextafter(t_old, direction * np.inf) - t_old)
        min_step   = next_after

        # Look for over/undershoots in previous step size
        if step_size > max_step:
            step_size = max_step
        elif step_size < min_step:
            step_size = min_step

        # Determine new step size
        step_accepted = False
        step_rejected = False
        step_error = False
        # # Step Loop
        while not step_accepted:

            if step_size < min_step:
                step_error = True
                break

            # Move time forward for this particular step size
            step = step_size * direction
            t_new = t_old + step

            # Check that we are not at the end of integration with that move
            if direction * (t_new - t_end) > 0:
                t_new = t_end

                # Correct the step if we were at the end of integration
                step = t_new - t_old
                step_size = np.abs(step)

            # Calculate derivative using RK method
            # Dot Product (K, a) * step
            for s in range(1, len_C):
                time_ = t_old + C[s] * step

                # Dot Product (K, a) * step
                if s == 1:
                    for i in range(y_size):
                        # Set the first column of K
                        temp_double_numeric = dydt_old[i]
                        K[0, i] = temp_double_numeric

                        # Calculate y_new for s==1
                        y_new[i] = y_old[i] + (temp_double_numeric * A_at_10 * step)
                else:
                    for j in range(s):
                        A_at_sj = A[s, j] * step
                        for i in range(y_size):
                            if j == 0:
                                # Initialize
                                y_new[i] = y_old[i]

                            y_new[i] += K[j, i] * A_at_sj

                # Update K with a new result from the differential equationC
                diffeq_output = np.asarray(diffeq(time_, y_new, *args), dtype=dtype)
                for i in range(y_size):
                    K[s, i] = diffeq_output[i]

            # Dot Product (K, B) * step
            for j in range(rk_n_stages):
                temp = B[j] * step
                # We do not use rk_n_stages_plus1 here because we are chopping off the last row of K to match
                #  the shape of B.
                for i in range(y_size):
                    if j == 0:
                        # Initialize
                        y_new[i] = y_old[i]
                    y_new[i] += K[j, i] * temp

            # Find final dydt for this timestep
            diffeq_output = np.asarray(diffeq(t_new, y_new, *args), dtype=dtype)

            for i in range(store_loop_size):
                if i < extra_start:
                    # Set diffeq results
                    dydt_new[i] = diffeq_output[i]
                else:
                    # Set extra results
                    extra_result[i - extra_start] = diffeq_output[i]

            # Estimate error to change the step size for next step
            if rk_method == 2:
                # DOP853 error estimation
                # Dot Product (K, E5) / Scale and (K, E3) / scale
                error_norm3 = 0.
                error_norm5 = 0.
                for i in range(y_size):
                    # Check how well this step performed
                    scale = atol_array[i] + np.maximum(np.abs(y_old[i]), np.abs(y_new[i])) * rtol_array[i]

                    # Set last array of K equal to dydt
                    K[rk_n_stages, i] = dydt_new[i]
                    for j in range(rk_n_stages_plus1):
                        if j == 0:
                            # Initialize
                            error_dot_1 = 0.
                            error_dot_2 = 0.

                        temp = K[j, i]
                        error_dot_1 += temp * E3[j]
                        error_dot_2 += temp * E5[j]

                    error_norm3_abs = np.abs(error_dot_1) / scale
                    error_norm5_abs = np.abs(error_dot_2) / scale

                    error_norm3 += (error_norm3_abs * error_norm3_abs)
                    error_norm5 += (error_norm5_abs * error_norm5_abs)

                # Check if errors are zero
                if (error_norm5 == 0.) and (error_norm3 == 0.):
                    error_norm = 0.
                else:
                    error_denom = error_norm5 + 0.01 * error_norm3
                    error_norm = step_size * error_norm5 / np.sqrt(error_denom * y_size)

            else:
                # Calculate Error for RK23 and RK45
                error_norm = 0.
                # Dot Product (K, E) * step / scale
                for i in range(y_size):
                    # Check how well this step performed.
                    scale = atol_array[i] + max(np.abs(y_old[i]), np.abs(y_new[i])) * rtol_array[i]

                    # Set last array of K equal to dydt
                    K[rk_n_stages, i] = dydt_new[i]
                    for j in range(rk_n_stages_plus1):
                        if j == 0:
                            # Initialize
                            error_dot_1 = 0.

                        error_dot_1 += K[j, i] * E[j]

                    error_norm_abs = np.abs(error_dot_1)
                    error_norm_abs *= (step / scale)

                    error_norm += (error_norm_abs * error_norm_abs)
                error_norm = np.sqrt(error_norm) / y_size_sqrt

            if error_norm < 1.:
                # The error is low! Let's update this step for the next time loop
                if error_norm == 0.:
                    step_factor = MAX_FACTOR
                else:
                    step_factor = min(
                            MAX_FACTOR,
                            SAFETY * error_norm**-error_expo
                            )

                if step_rejected:
                    # There were problems with this step size on the previous step loop. Make sure factor does
                    #    not exasperate them.
                    step_factor = min(step_factor, 1.)

                step_size = step_size * step_factor
                step_accepted = True
            else:
                step_size = step_size * max(
                        MIN_FACTOR,
                        SAFETY * error_norm**-error_expo
                        )
                step_rejected = True

        if step_error:
            # Issue with step convergence
            status = -1
            message = "Error in step size calculation:\n\tRequired step size is less than spacing between numbers."
            break
        elif not step_accepted:
            # Issue with step convergence
            status = -7
            message = "Error in step size calculation:\n\tError in step size acceptance."
            break

        # End of step loop. Update the _now variables
        t_old = t_new

        for i in range(y_size):
            y_old[i] = y_new[i]
            dydt_old[i] = dydt_new[i]

        # Save data
        # If there is extra outputs then we need to store those at this timestep as well.
        for i in range(store_loop_size):
            if i < extra_start:
                # Pull from y result
                y_result_store[i] = y_new[i]
            else:
                # Pull from extra
                y_result_store[i] = extra_result[i - extra_start]

        time_domain_list.append(t_new)
        y_result_list.append(np.copy(y_result_store))
        len_t += 1

    # To match the format that scipy follows, we will take the transpose of y.
    time_domain = np.empty(len_t, dtype=np.float64)
    y_results = np.empty((store_loop_size, len_t), dtype=dtype)
    for t_i in range(len_t):
        time_domain[t_i] = time_domain_list[t_i]
        y_results_list_at_t = y_result_list[t_i]
        for y_i in range(store_loop_size):
            y_results[y_i, t_i] = y_results_list_at_t[y_i]

    success = False
    if status == 1:
        success = True
        message = "Integration completed without issue."

    if t_eval_size > 0:
        old_status = status
        status = 2
        # User only wants data at specific points.
        # The current version of this function has not implemented sicpy's dense output, so we must use an interpolation.
        t_eval = np.asarray(t_eval, dtype=np.float64)
        y_results_reduced = np.empty((total_size, t_eval.size), dtype=dtype)

        # np.interp only works on 1D arrays, so we have to loop through each of the variables and call for each y:
        for i in range(y_size):
            y_results_reduced[i, :] = np.interp(t_eval, time_domain, y_results[i, :])

        if capture_extra:
            # Right now if there is any extra output then it is stored at each time step used in the RK loop.
            # We have to make a choice on what to output do we, like we do with y, interpolate all of those extras?
            #  or do we use the interpolation on y to find new values.
            # The latter method is more computationally expensive (recalls the diffeq for each y) but is more accurate.
            if interpolate_extra:
                # Continue the interpolation for the extra values.
                for i in range(extra_size):
                    y_results_reduced[extra_start + i, :] = \
                        np.interp(t_eval, time_domain, y_results[extra_start + i, :])
            else:
                # Use y and t to recalculate the extra outputs
                y_ = np.empty(y_size, dtype=dtype)
                for t_i in range(t_eval_size):
                    t_ = t_eval[t_i]
                    for y_i in range(y_size):
                        y_[y_i] = y_results_reduced[y_i, t_i]
                    diffeq_output = np.asarray(diffeq(t_, y_, *args), dtype=dtype)
                    for i in range(extra_size):
                        y_results_reduced[extra_start + i, t_i] = diffeq_output[extra_start + i]

        y_results = y_results_reduced
        time_domain = t_eval

        status = old_status

    # Install result into solution named tuple
    nbrk_result = WrapNBRKResult(success, message, time_domain.size, total_size, y0.size, status, y_results, time_domain)

    return nbrk_result



# distutils: language = c++
# cython: boundscheck=False, wraparound=False, nonecheck=False, cdivision=True, initializedcheck=False
from libc.math cimport floor

import numpy as np
cimport numpy as cnp
cnp.import_array()

# =====================================================================================================================
# Import CySolverResult (container for integration results)
# =====================================================================================================================
cdef class WrapCySolverResult:

    cdef void set_cyresult_pointer(self, shared_ptr[CySolverResult] cyresult_shptr):

        # Store c++ based result and pull out key information
        self.cyresult_shptr = cyresult_shptr
        self.cyresult_ptr   = cyresult_shptr.get()

        # Convert solution to pointers and views
        if self.cyresult_ptr.size > 0:
            self.time_ptr  = &self.cyresult_ptr.time_domain_vec[0]
            self.y_ptr     = &self.cyresult_ptr.solution[0]
            self.time_view = <double[:self.size]>self.time_ptr
            self.y_view    = <double[:self.size * self.num_dy]>self.y_ptr
    
    def call(self, double t):
        """ Call the dense output interpolater and return y """

        if not self.cyresult_ptr.capture_dense_output:
            raise AttributeError("Can not call WrapCySolverResult when dense_output set to False.")

        y_interp_array = np.empty(self.cyresult_ptr.num_dy, dtype=np.float64, order='C')
        cdef double[::1] y_interp_view = y_interp_array
        cdef double* y_interp_ptr      = &y_interp_view[0]

        self.cyresult_ptr.call(t, y_interp_ptr)
        return y_interp_array
    
    def call_vectorize(self, double[::1] t_view):
        """ Call the dense output interpolater and return y """

        if not self.cyresult_ptr.capture_dense_output:
            raise AttributeError("Can not call WrapCySolverResult when dense_output set to False.")

        cdef size_t len_t = len(t_view)

        y_interp_array = np.empty(self.cyresult_ptr.num_dy * len_t, dtype=np.float64, order='C')
        cdef double[::1] y_interp_view = y_interp_array
        cdef double* y_interp_ptr = &y_interp_view[0]
        cdef double* t_array_ptr  = &t_view[0]

        self.cyresult_ptr.call_vectorize(t_array_ptr, len_t, y_interp_ptr)
        return y_interp_array.reshape(len_t, self.cyresult_ptr.num_dy).T

    def print_diagnostics(self):

        cdef str diagnostic_str = ''
        from CyRK import __version__
        cdef str direction_str = 'Forward'
        if self.cyresult_ptr.direction_flag == 0:
            direction_str = 'Backward'
        
        diagnostic_str += f'CyRK v{__version__} - WrapCySolverResult Diagnostic.\n'
        diagnostic_str += f'\n----------------------------------------------------\n'
        diagnostic_str += f'# of y:      {self.num_y}.\n'
        diagnostic_str += f'# of dy:     {self.num_dy}.\n'
        diagnostic_str += f'Success:     {self.success}.\n'
        diagnostic_str += f'Error Code:  {self.error_code}.\n'
        diagnostic_str += f'Size:        {self.size}.\n'
        diagnostic_str += f'Steps Taken: {self.steps_taken}.\n'
        diagnostic_str += f'Message:\n\t{self.message}\n'
        diagnostic_str += f'\n----------------- CySolverResult -------------------\n'
        diagnostic_str += f'Capture Extra:         {self.cyresult_ptr.capture_extra}.\n'
        diagnostic_str += f'Capture Dense Output:  {self.cyresult_ptr.capture_dense_output}.\n'
        diagnostic_str += f'Integration Direction: {direction_str}.\n'
        diagnostic_str += f'Integration Method:    {self.cyresult_ptr.integrator_method}.\n'
        diagnostic_str += f'# of Interpolates:     {self.cyresult_ptr.num_interpolates}.\n'

        cdef CySolverBase* cysolver = self.cyresult_ptr.solver_uptr.get()
        cdef size_t num_y
        cdef size_t num_dy
        cdef size_t i
        cdef size_t args_size
        cdef size_t args_size_dbls
        cdef double* args_dbl_ptr
        if cysolver:
            num_y  = cysolver.num_y
            num_dy = cysolver.num_dy
            diagnostic_str += f'\n------------------ CySolverBase --------------------\n'
            diagnostic_str += f'Status:     {cysolver.status}.\n'
            diagnostic_str += f'# of y:     {num_y}.\n'
            diagnostic_str += f'# of dy:    {num_dy}.\n'
            diagnostic_str += f'PySolver:   {cysolver.use_pysolver}.\n'
            diagnostic_str += f't_now:      {cysolver.t_now}.\n'
            diagnostic_str += f'y_now:\n'
            for i in range(num_y):
                diagnostic_str += f'\ty{i}  = {cysolver.y_now[i]:0.5e}.\n'
            diagnostic_str += f'dy_now:\n'
            for i in range(num_dy):
                diagnostic_str += f'\tdy{i} = {cysolver.dy_now[i]:0.5e}.\n'
            args_size      = cysolver.size_of_args
            args_size_dbls = <size_t>floor(args_size / sizeof(double))
            args_dbl_ptr   = <double*>cysolver.args_ptr
            diagnostic_str += f'args size (bytes):   {args_size}.\n'
            diagnostic_str += f'args size (doubles): {args_size_dbls}.\n'
            if args_size_dbls > 0:
                diagnostic_str += f'args (as doubles):\n'
                for i in range(args_size_dbls):
                    diagnostic_str += f'\targ{i} = {args_dbl_ptr[i]:0.5e}.\n'
        else:
            diagnostic_str += f'CySolverBase instance was deleted or voided.\n'

        diagnostic_str += f'\n-------------- Diagnostic Complete -----------------\n'
        print(diagnostic_str)

    def __dealloc__(self):
        # Deallocate any heap allocated memory
        self.cyresult_shptr.reset()

    @property
    def success(self):
        return self.cyresult_ptr.success
        
    @property
    def message(self):
        return str(self.cyresult_ptr.message_ptr, 'UTF-8')
    
    @property
    def t(self):
        return np.copy(np.asarray(self.time_view, dtype=np.float64, order='C'))
    
    @property
    def y(self):
        return np.copy(np.asarray(self.y_view, dtype=np.float64, order='C')).reshape((self.size, self.num_dy)).T
    
    @property
    def size(self):
        return self.cyresult_ptr.size
    
    @property
    def steps_taken(self):
        return self.cyresult_ptr.steps_taken
    
    @property
    def num_y(self):
        return self.cyresult_ptr.num_y
    
    @property
    def num_dy(self):
        return self.cyresult_ptr.num_dy
    
    @property
    def error_code(self):
        return self.cyresult_ptr.error_code
    
    def __call__(self, t):

        if type(t) == cnp.ndarray:
            return self.call_vectorize(t)
        else:
            return self.call(t).reshape(self.cyresult_ptr.num_dy, 1)


# =====================================================================================================================
# Create Wrapped cysolve_ivp (has various defaults)
# =====================================================================================================================
cdef void cysolve_ivp_noreturn(
            shared_ptr[CySolverResult] solution_sptr,
            DiffeqFuncType diffeq_ptr,
            const double* t_span_ptr,
            const double* y0_ptr,
            const size_t num_y,
            int method = 1,
            double rtol = 1.0e-3,
            double atol = 1.0e-6,
            char* args_ptr = NULL,
            size_t size_of_args = 0,
            size_t num_extra = 0,
            size_t max_num_steps = 0,
            size_t max_ram_MB = 2000,
            bint dense_output = False,
            double* t_eval = NULL,
            size_t len_t_eval = 0,
            PreEvalFunc pre_eval_func = NULL,
            double* rtols_ptr = NULL,
            double* atols_ptr = NULL,
            double max_step = MAX_STEP,
            double first_step = 0.0,
            size_t expected_size = 0
            ) noexcept nogil:
    baseline_cysolve_ivp_noreturn(
        solution_sptr,
        diffeq_ptr,
        t_span_ptr,
        y0_ptr,
        num_y,
        method,
        expected_size,
        num_extra,
        args_ptr,
        size_of_args,
        max_num_steps,
        max_ram_MB,
        dense_output,
        t_eval,
        len_t_eval,
        pre_eval_func,
        rtol,
        atol,
        rtols_ptr,
        atols_ptr,
        max_step,
        first_step
        )

cdef CySolveOutput cysolve_ivp(
            DiffeqFuncType diffeq_ptr,
            const double* t_span_ptr,
            const double* y0_ptr,
            const size_t num_y,
            int method = 1,
            double rtol = 1.0e-3,
            double atol = 1.0e-6,
            char* args_ptr = NULL,
            size_t size_of_args = 0,
            size_t num_extra = 0,
            size_t max_num_steps = 0,
            size_t max_ram_MB = 2000,
            bint dense_output = False,
            double* t_eval = NULL,
            size_t len_t_eval = 0,
            PreEvalFunc pre_eval_func = NULL,
            double* rtols_ptr = NULL,
            double* atols_ptr = NULL,
            double max_step = MAX_STEP,
            double first_step = 0.0,
            size_t expected_size = 0
            ) noexcept nogil:
    cdef CySolveOutput result = baseline_cysolve_ivp(
        diffeq_ptr,
        t_span_ptr,
        y0_ptr,
        num_y,
        method,
        expected_size,
        num_extra,
        args_ptr,
        size_of_args,
        max_num_steps,
        max_ram_MB,
        dense_output,
        t_eval,
        len_t_eval,
        pre_eval_func,
        rtol,
        atol,
        rtols_ptr,
        atols_ptr,
        max_step,
        first_step
        )

    return result

cdef CySolveOutput cysolve_ivp_gil(
            DiffeqFuncType diffeq_ptr,
            const double* t_span_ptr,
            const double* y0_ptr,
            const size_t num_y,
            int method = 1,
            double rtol = 1.0e-3,
            double atol = 1.0e-6,
            char* args_ptr = NULL,
            size_t size_of_args = 0,
            size_t num_extra = 0,
            size_t max_num_steps = 0,
            size_t max_ram_MB = 2000,
            bint dense_output = False,
            double* t_eval = NULL,
            size_t len_t_eval = 0,
            PreEvalFunc pre_eval_func = NULL,
            double* rtols_ptr = NULL,
            double* atols_ptr = NULL,
            double max_step = MAX_STEP,
            double first_step = 0.0,
            size_t expected_size = 0
            ) noexcept:
    
    cdef CySolveOutput result = baseline_cysolve_ivp(
        diffeq_ptr,
        t_span_ptr,
        y0_ptr,
        num_y,
        method,
        expected_size,
        num_extra,
        args_ptr,
        size_of_args,
        max_num_steps,
        max_ram_MB,
        dense_output,
        t_eval,
        len_t_eval,
        pre_eval_func,
        rtol,
        atol,
        rtols_ptr,
        atols_ptr,
        max_step,
        first_step
        )

    return result



from libcpp cimport bool as cpp_bool
from libcpp.vector cimport vector
from libcpp.memory cimport shared_ptr, unique_ptr

cimport cpython.ref as cpy_ref
from CyRK.cy.pysolver_cyhook cimport DiffeqMethod

cimport numpy as cnp
cnp.import_array()


# =====================================================================================================================
# Import common functions and constants
# =====================================================================================================================
cdef extern from "common.cpp" nogil:
    const double INF
    const double EPS_100
    const size_t BUFFER_SIZE
    const double MAX_STEP

    ctypedef void (*PreEvalFunc)(char*, double, double*, char*)
    ctypedef void (*DiffeqFuncType)(double*, double, double*, char*, PreEvalFunc)

    cdef size_t find_expected_size(        
        size_t num_y,
        size_t num_extra,
        double t_delta_abs,
        double rtol_min)


cdef extern from "cy_array.cpp" nogil:
    size_t binary_search_with_guess(double key, const double* array, size_t length, size_t guess)


cdef extern from "dense.cpp" nogil:
    cdef cppclass CySolverDense:
        CySolverDense()
        CySolverDense(
            int integrator_int,
            CySolverBase* solver_ptr,
            cpp_bool set_state)

        int integrator_int
        size_t Q_order
        size_t num_y
        CySolverBase* solver_ptr
        double t_old
        double t_now
        double step

        vector[double] state_data_vec

        void set_state()
        void call(double t_interp, double* y_interped)


# =====================================================================================================================
# Import CySolverResult (container for integration results)
# =====================================================================================================================
cdef extern from "cysolution.cpp" nogil:
    cdef cppclass CySolverResult:
            CySolverResult()
            CySolverResult(
                const size_t num_y,
                const size_t num_extra,
                const size_t expected_size,
                const size_t last_t,
                const cpp_bool direction_flag,
                const cpp_bool capture_dense_output,
                const cpp_bool t_eval_provided)
            
            cpp_bool capture_extra
            cpp_bool retain_solver
            cpp_bool capture_dense_output
            cpp_bool t_eval_provided
            cpp_bool success
            cpp_bool reset_called
            cpp_bool solver_reset_called
            cpp_bool direction_flag
            int error_code
            int integrator_method
            size_t num_y
            size_t num_extra
            size_t num_dy
            char* message_ptr
            size_t size
            size_t num_interpolates
            size_t steps_taken
            vector[double] time_domain_vec
            vector[double] time_domain_vec_sorted
            vector[double] solution
            vector[double]* time_domain_vec_sorted_ptr
            vector[CySolverDense] dense_vec
            unique_ptr[CySolverBase] solver_uptr
            vector[double] interp_time_vec

            void save_data(double new_t, double* new_solution_y, double* new_solution_dy)
            int build_dense(cpp_bool save)
            void solve()
            void finalize()
            void reset()
            void reset_solver()
            void build_solver(
                DiffeqFuncType diffeq_ptr,
                const double t_start,
                const double t_end,
                const double* y0_ptr,
                const int method,
                const size_t expected_size,
                const char* args_ptr,
                const size_t size_of_args,
                const size_t max_num_steps,
                const size_t max_ram_MB,
                const double* t_eval,
                const size_t len_t_eval,
                PreEvalFunc pre_eval_func,
                const double rtol,
                const double atol,
                const double* rtols_ptr,
                const double* atols_ptr,
                const double max_step_size,
                const double first_step_size
            )
            void update_message(const char* new_message)
            void call(const double t, double* y_interp)
            void call_vectorize(const double* t_array_ptr, size_t len_t, double* y_interp)

ctypedef shared_ptr[CySolverResult] CySolveOutput

cdef class WrapCySolverResult:
    """ Wrapper for the C++ class `CySolverResult` defined in "cysolution.cpp" """

    cdef shared_ptr[CySolverResult] cyresult_shptr
    cdef CySolverResult* cyresult_ptr
    cdef double* time_ptr
    cdef double* y_ptr
    cdef double[::1] time_view
    cdef double[::1] y_view

    cdef void set_cyresult_pointer(self, shared_ptr[CySolverResult] cyresult_shptr)


# =====================================================================================================================
# Import CySolver Integrator Base Class
# =====================================================================================================================
cdef extern from "cysolver.cpp" nogil:
    struct NowStatePointers:
        double* t_now_ptr
        double* y_now_ptr
        double* dy_now_ptr

    cdef cppclass CySolverBase:
        CySolverBase()
        CySolverBase(
            DiffeqFuncType diffeq_ptr,
            CySolverResult* storage_ptr,
            const double t_start,
            const double t_end,
            const double* y0_ptr,
            const size_t num_y,
            const size_t num_extra,
            const char* args_ptr,
            const size_t size_of_args,
            const size_t max_num_steps,
            const size_t max_ram_MB,
            const cpp_bool use_dense_output,
            const double* t_eval,
            const size_t len_t_eval,
            PreEvalFunc pre_eval_func
        )

        cpp_bool use_pysolver
        DiffeqMethod py_diffeq_method
        cpy_ref.PyObject cython_extension_class_instance
        int status
        int integration_method
        size_t num_dy
        size_t num_y
        CySolverResult* storage_ptr
        size_t size_of_args
        vector[char] args_char_vec
        char* args_ptr
        size_t len_t
        double t_now
        vector[double] y_now
        vector[double] dy_now

        cpp_bool check_status()
        NowStatePointers get_now_state()
        void reset()
        void offload_to_temp()
        void load_back_from_temp()
        void calc_first_step_size()
        void take_step()
        void solve()
        void cy_diffeq()
        void diffeq()
        void set_cython_extension_instance(cpy_ref.PyObject* cython_extension_class_instance)
        void py_diffeq()


# =====================================================================================================================
# Import CySolver Runge-Kutta Integrators
# =====================================================================================================================
cdef extern from "rk.cpp" nogil:
    cdef cppclass RKSolver(CySolverBase):
        RKSolver()
        RKSolver(
            DiffeqFuncType diffeq_ptr,
            CySolverResult* storage_ptr,
            const double t_start,
            const double t_end,
            const double* y0_ptr,
            const size_t num_y,
            const size_t num_extra,
            const char* args_ptr,
            const size_t size_of_args,
            const size_t max_num_steps,
            const size_t max_ram_MB,
            const cpp_bool use_dense_output,
            const double* t_eval,
            const size_t len_t_eval,
            PreEvalFunc pre_eval_func,
            const double rtol,
            const double atol,
            const double* rtols_ptr,
            const double* atols_ptr,
            const double max_step_size,
            const double first_step_size
        )
        void p_step_implementation()
        void p_estimate_error()
        void reset()
        void calc_first_step_size()


    cdef int RK23_METHOD_INT
    cdef cppclass RK23(RKSolver):
        RK23()
        RK23(
            DiffeqFuncType diffeq_ptr,
            CySolverResult* storage_ptr,
            const double t_start,
            const double t_end,
            const double* y0_ptr,
            const size_t num_y,
            const size_t num_extra,
            const char* args_ptr,
            const size_t size_of_args,
            const size_t max_num_steps,
            const size_t max_ram_MB,
            const cpp_bool use_dense_output,
            const double* t_eval,
            const size_t len_t_eval,
            PreEvalFunc pre_eval_func,
            const double rtol,
            const double atol,
            const double* rtols_ptr,
            const double* atols_ptr,
            const double max_step_size,
            const double first_step_size
        )
        void reset()
    
    cdef int RK45_METHOD_INT
    cdef cppclass RK45(RKSolver):
        RK45()
        RK45(
            DiffeqFuncType diffeq_ptr,
            CySolverResult* storage_ptr,
            const double t_start,
            const double t_end,
            const double* y0_ptr,
            const size_t num_y,
            const size_t num_extra,
            const char* args_ptr,
            const size_t size_of_args,
            const size_t max_num_steps,
            const size_t max_ram_MB,
            const cpp_bool use_dense_output,
            const double* t_eval,
            const size_t len_t_eval,
            PreEvalFunc pre_eval_func,
            const double rtol,
            const double atol,
            const double* rtols_ptr,
            const double* atols_ptr,
            const double max_step_size,
            const double first_step_size
        )
        void reset()
    
    cdef int DOP853_METHOD_INT
    cdef cppclass DOP853(RKSolver):
        DOP853()
        DOP853(
            DiffeqFuncType diffeq_ptr,
            CySolverResult* storage_ptr,
            const double t_start,
            const double t_end,
            const double* y0_ptr,
            const size_t num_y,
            const size_t num_extra,
            const char* args_ptr,
            const size_t size_of_args,
            const size_t max_num_steps,
            const size_t max_ram_MB,
            const cpp_bool use_dense_output,
            const double* t_eval,
            const size_t len_t_eval,
            PreEvalFunc pre_eval_func,
            const double rtol,
            const double atol,
            const double* rtols_ptr,
            const double* atols_ptr,
            const double max_step_size,
            const double first_step_size
        )
        void reset()


# =====================================================================================================================
# Import the C++ cysolve_ivp helper function
# =====================================================================================================================
cdef extern from "cysolve.cpp" nogil:
    # Pure C++ and Cython implementation

    cdef void baseline_cysolve_ivp_noreturn(
            shared_ptr[CySolverResult] solution_sptr,
            DiffeqFuncType diffeq_ptr,
            const double* t_span_ptr,
            const double* y0_ptr,
            const size_t num_y,
            const int method,
            const size_t expected_size,
            const size_t num_extra,
            const char* args_ptr,
            const size_t size_of_args,
            const size_t max_num_steps,
            const size_t max_ram_MB,
            const cpp_bool dense_output,
            const double* t_eval,
            const size_t len_t_eval,
            PreEvalFunc pre_eval_func,
            const double rtol,
            const double atol,
            const double* rtols_ptr,
            const double* atols_ptr,
            const double max_step_size,
            const double first_step_size
            )

    cdef shared_ptr[CySolverResult] baseline_cysolve_ivp(
            DiffeqFuncType diffeq_ptr,
            const double* t_span_ptr,
            const double* y0_ptr,
            const size_t num_y,
            const int method,
            const size_t expected_size,
            const size_t num_extra,
            const char* args_ptr,
            const size_t size_of_args,
            const size_t max_num_steps,
            const size_t max_ram_MB,
            const cpp_bool dense_output,
            const double* t_eval,
            const size_t len_t_eval,
            PreEvalFunc pre_eval_func,
            const double rtol,
            const double atol,
            const double* rtols_ptr,
            const double* atols_ptr,
            const double max_step_size,
            const double first_step_size
            )


# =====================================================================================================================
# Cython-based wrapper for baseline_cysolve_ivp that carries default values.
# =====================================================================================================================
cdef void cysolve_ivp_noreturn(
    shared_ptr[CySolverResult] solution_sptr,
    DiffeqFuncType diffeq_ptr,
    const double* t_span_ptr,
    const double* y0_ptr,
    const size_t num_y,
    int method = *,
    double rtol = *,
    double atol = *,
    char* args_ptr = *,
    size_t size_of_args = *,
    size_t num_extra = *,
    size_t max_num_steps = *,
    size_t max_ram_MB = *,
    bint dense_output = *,
    double* t_eval = *,
    size_t len_t_eval = *,
    PreEvalFunc pre_eval_func = *,
    double* rtols_ptr = *,
    double* atols_ptr = *,
    double max_step = *,
    double first_step = *,
    size_t expected_size = *
    ) noexcept nogil

cdef CySolveOutput cysolve_ivp(
    DiffeqFuncType diffeq_ptr,
    const double* t_span_ptr,
    const double* y0_ptr,
    const size_t num_y,
    int method = *,
    double rtol = *,
    double atol = *,
    char* args_ptr = *,
    size_t size_of_args = *,
    size_t num_extra = *,
    size_t max_num_steps = *,
    size_t max_ram_MB = *,
    bint dense_output = *,
    double* t_eval = *,
    size_t len_t_eval = *,
    PreEvalFunc pre_eval_func = *,
    double* rtols_ptr = *,
    double* atols_ptr = *,
    double max_step = *,
    double first_step = *,
    size_t expected_size = *
    ) noexcept nogil

cdef CySolveOutput cysolve_ivp_gil(
    DiffeqFuncType diffeq_ptr,
    const double* t_span_ptr,
    const double* y0_ptr,
    const size_t num_y,
    int method = *,
    double rtol = *,
    double atol = *,
    char* args_ptr = *,
    size_t size_of_args = *,
    size_t num_extra = *,
    size_t max_num_steps = *,
    size_t max_ram_MB = *,
    bint dense_output = *,
    double* t_eval = *,
    size_t len_t_eval = *,
    PreEvalFunc pre_eval_func = *,
    double* rtols_ptr = *,
    double* atols_ptr = *,
    double max_step = *,
    double first_step = *,
    size_t expected_size = *
    ) noexcept



""" Helper function to convert conventional solve_ivp or numba diffeq's into a format that cyrk can accept. """
import numpy as np
from numba import njit

def nb2cy(diffeq: callable, use_njit: bool = True, cache_njit: bool = False) -> callable:
    """ Convert numba/scipy differential equation functions to the cyrk format.

    Parameters
    ----------
    diffeq : callable
        Differential equation function.
    use_njit : bool = True
        If True, the final function will be njited.
    cache_njit : bool = False
        If True, then the njit-complied function will be cached.

    Returns
    -------
    diffeq_cyrk : callable
        cyrk-safe differential equation function.
    """


    if use_njit:
        if cache_njit:
            njit_ = njit(cache=True)
        else:
            njit_ = njit(cache=False)
    else:
        def njit_(func):
            return func

    @njit_
    def diffeq_cyrk(dy, t, y, *args):
        # Cython integrator requires the arguments to be passed as input args
        dy_ = diffeq(t, y, *args)

        # Set the input dy items equal to the output
        for i in range(y.size):
            dy[i] = dy_[i]

    return diffeq_cyrk


def cy2nb(diffeq: callable, use_njit: bool = True, cache_njit: bool = False) -> callable:
    """ Convert cyrk differential equation functions to the numba/scipy format.

    Parameters
    ----------
    diffeq : callable
        Differential equation function.
    use_njit : bool = True
        If True, the final function will be njited.
    cache_njit : bool = False
        If True, then the njit-complied function will be cached.

    Returns
    -------
    diffeq_nbrk : callable
        numba/scipy-safe differential equation function.
    """

    if use_njit:
        if cache_njit:
            njit_ = njit(cache=True)
        else:
            njit_ = njit(cache=False)
    else:
        def njit_(func):
            return func

    @njit_
    def diffeq_nbrk(t, y, *args):
        # Cython integrator requires the arguments to be passed as input args
        dy = np.empty_like(y)
        diffeq(dy, t, y, *args)

        return dy

    return diffeq_nbrk


from collections import namedtuple

WrapNBRKResult = namedtuple(
     'WrapNBRKResult',
        (
            'success',
            'message',
            'size',
            'num_dy',
            'num_y',
            'error_code',
            'y',
            't'
        )
    )



# distutils: language = c++
# cython: boundscheck=False, wraparound=False, nonecheck=False, cdivision=True, initializedcheck=False

from libcpp cimport bool as cpp_bool
from libcpp.limits cimport numeric_limits
from libc.math cimport sin, cos
from libc.stdlib cimport malloc, free, realloc

cdef double d_NAN = numeric_limits[double].quiet_NaN()

from CyRK.cy.cysolver_api cimport cysolve_ivp, WrapCySolverResult, DiffeqFuncType,MAX_STEP, CySolveOutput

import numpy as np
cimport numpy as np
np.import_array()


cdef void baseline_diffeq(double* dy_ptr, double t, double* y_ptr, char* args_ptr, PreEvalFunc pre_eval_func) noexcept nogil:
    # Unpack y
    cdef double y0, y1
    y0 = y_ptr[0]
    y1 = y_ptr[1]

    dy_ptr[0] = (1. - 0.01 * y1) * y0
    dy_ptr[1] = (0.02 * y0 - 1.) * y1


cdef void accuracy_test_diffeq(double* dy_ptr, double t, double* y_ptr, char* args_ptr, PreEvalFunc pre_eval_func) noexcept nogil:
    # Unpack y
    cdef double y0, y1
    y0 = y_ptr[0]
    y1 = y_ptr[1]

    dy_ptr[0] = sin(t) - y1
    dy_ptr[1] = cos(t) + y0


cdef void extraoutput_test_diffeq(double* dy_ptr, double t, double* y_ptr, char* args_ptr, PreEvalFunc pre_eval_func) noexcept nogil:
    # Unpack y
    cdef double y0, y1, extra_0, extra_1
    y0 = y_ptr[0]
    y1 = y_ptr[1]

    extra_0 = (1. - 0.01 * y1)
    extra_1 = (0.02 * y0 - 1.)

    # Store dy/dt
    dy_ptr[0] = extra_0 * y0
    dy_ptr[1] = extra_1 * y1

    # Store extra output
    dy_ptr[2] = extra_0
    dy_ptr[3] = extra_1


cdef void lorenz_diffeq(double* dy_ptr, double t, double* y_ptr, char* args_ptr, PreEvalFunc pre_eval_func) noexcept nogil:
    # Unpack args
    cdef double* args_dbl_ptr = <double*>args_ptr
    cdef double a = args_dbl_ptr[0]
    cdef double b = args_dbl_ptr[1]
    cdef double c = args_dbl_ptr[2]

    # Unpack y
    cdef double y0, y1, y2
    y0 = y_ptr[0]
    y1 = y_ptr[1]
    y2 = y_ptr[2]

    dy_ptr[0] = a * (y1 - y0)
    dy_ptr[1] = y0 * (b - y2) - y1
    dy_ptr[2] = y0 * y1 - c * y2


cdef void lorenz_extraoutput_diffeq(double* dy_ptr, double t, double* y_ptr, char* args_ptr, PreEvalFunc pre_eval_func) noexcept nogil:
    # Unpack args
    cdef double* args_dbl_ptr = <double*>args_ptr
    cdef double a = args_dbl_ptr[0]
    cdef double b = args_dbl_ptr[1]
    cdef double c = args_dbl_ptr[2]

    # Unpack y
    cdef double y0, y1, y2
    y0 = y_ptr[0]
    y1 = y_ptr[1]
    y2 = y_ptr[2]

    cdef double e_1 = a
    cdef double e_2 = (b - y2)
    cdef double e_3 = c * y2

    dy_ptr[0] = e_1 * (y1 - y0)
    dy_ptr[1] = y0 * e_2 - y1
    dy_ptr[2] = y0 * y1 - e_3

    dy_ptr[3] = e_1
    dy_ptr[4] = e_2
    dy_ptr[5] = e_3


cdef void lotkavolterra_diffeq(double* dy_ptr, double t, double* y_ptr, char* args_ptr, PreEvalFunc pre_eval_func) noexcept nogil:
    # Unpack args
    cdef double* args_dbl_ptr = <double*>args_ptr
    cdef double a = args_dbl_ptr[0]
    cdef double b = args_dbl_ptr[1]
    cdef double c = args_dbl_ptr[2]
    cdef double d = args_dbl_ptr[3]

    # Unpack y
    cdef double y0, y1
    y0 = y_ptr[0]
    y1 = y_ptr[1]

    dy_ptr[0] = a * y0 - b * y0 * y1
    dy_ptr[1] = -c * y1 + d * y0 * y1



cdef void pendulum_diffeq(double* dy_ptr, double t, double* y_ptr, char* args_ptr, PreEvalFunc pre_eval_func) noexcept nogil:
    # Unpack args
    cdef double* args_dbl_ptr = <double*>args_ptr
    cdef double l = args_dbl_ptr[0]
    cdef double m = args_dbl_ptr[1]
    cdef double g = args_dbl_ptr[2]

    cdef double coeff_1 = (-3. * g / (2. * l))
    cdef double coeff_2 = (3. / (m * l**2))

    # Unpack y
    cdef double y0, y1, torque
    y0 = y_ptr[0]
    y1 = y_ptr[1]

    # External torque
    torque = 0.1 * sin(t)

    dy_ptr[0] = y1
    dy_ptr[1] = coeff_1 * sin(y0) + coeff_2 * torque


cdef struct ArbitraryArgStruct:
    double l
    # Let's make sure it has something in the middle that is not a double for size checks.
    cpp_bool cause_fail
    cpp_bool checker
    double m
    double g

cdef void arbitrary_arg_test(double* dy_ptr, double t, double* y_ptr, char* args_ptr, PreEvalFunc pre_eval_func) noexcept nogil:
    # Unpack args
    cdef ArbitraryArgStruct* arb_args_ptr = <ArbitraryArgStruct*>args_ptr
    cdef double l = arb_args_ptr.l
    cdef double m = arb_args_ptr.m
    cdef double g = arb_args_ptr.g
    cdef cpp_bool cause_fail = arb_args_ptr.cause_fail
    cdef cpp_bool checker    = arb_args_ptr.checker

    cdef double coeff_1 = (-3. * g / (2. * l))
    cdef double coeff_2 = (3. / (m * l**2))

    # Unpack y
    cdef double y0, y1, torque
    y0 = y_ptr[0]
    y1 = y_ptr[1]

    # External torque
    torque = 0.1 * sin(t)

    if cause_fail:
        # Put something crazy that will cause problems
        dy_ptr[0] = (100 * t)**(1000)
        dy_ptr[1] = (-100 * t)**(1000)
    else:
        dy_ptr[0] = y1
        dy_ptr[1] = coeff_1 * sin(y0) + coeff_2 * torque



cdef void pendulum_preeval_func(char* output_ptr, double time, double* y_ptr, char* args_ptr) noexcept nogil:

    # Unpack args
    cdef double* args_dbl_ptr = <double*>args_ptr
    cdef double l = args_dbl_ptr[0]
    cdef double m = args_dbl_ptr[1]
    cdef double g = args_dbl_ptr[2]

    cdef double coeff_1 = (-3. * g / (2. * l))
    cdef double coeff_2 = (3. / (m * l**2))

    # Unpack y
    cdef double torque

    # External torque
    torque = 0.1 * sin(time)

    # Convert output pointer to double pointer so we can store data
    cdef double* output_dbl_ptr = <double*>output_ptr
    output_dbl_ptr[0] = torque
    output_dbl_ptr[1] = coeff_1
    output_dbl_ptr[2] = coeff_2


cdef void pendulum_preeval_diffeq(double* dy_ptr, double t, double* y_ptr, char* args_ptr, PreEvalFunc pre_eval_func) noexcept nogil:
    
    # Make stack allocated storage for pre eval output
    cdef double[3] pre_eval_storage
    cdef double* pre_eval_storage_ptr = &pre_eval_storage[0]

    # Cast storage to void so we can call function
    cdef char* pre_eval_storage_char_ptr = <char*>pre_eval_storage_ptr

    # Call Pre-Eval Function
    pre_eval_func(pre_eval_storage_char_ptr, t, y_ptr, args_ptr)

    cdef double y0 = y_ptr[0]
    cdef double y1 = y_ptr[1]

    dy_ptr[0] = y1
    dy_ptr[1] = pre_eval_storage_ptr[1] * sin(y0) + pre_eval_storage_ptr[2] * pre_eval_storage_ptr[0]


def cy_extra_output_tester():

    cdef double[2] t_span = [0., 10.]
    cdef double* t_span_ptr = &t_span[0]

    cdef double[3] y0 = [1., 0., 0]
    cdef double* y0_ptr = &y0[0]
    
    cdef size_t num_y     = 3
    cdef size_t num_extra = 3
    cdef int int_method   = 1

    cdef size_t arg_size = sizeof(double) * 3
    cdef double* args_ptr = <double*>malloc(arg_size)
    args_ptr[0] = 10.0
    args_ptr[1] = 28.0
    args_ptr[2] = 8.0 / 3.0

    cdef CySolveOutput result = cysolve_ivp(
        lorenz_extraoutput_diffeq,
        t_span_ptr,
        y0_ptr,
        num_y,
        method=int_method,
        rtol=1.0e-4,
        atol=1.0e-5,
        args_ptr=<char*>args_ptr,
        size_of_args=arg_size,
        num_extra=num_extra,
        max_num_steps=0,
        max_ram_MB=2000,
        dense_output=True
        )

    cdef double check_t = 4.335
    assert result.get().success

    # Call the result to get the baseline values for extra output
    cdef double[6] y_interp
    cdef double* y_interp_ptr = &y_interp[0]
    result.get().call(check_t, y_interp_ptr)
    cdef double dy1 = y_interp_ptr[0]
    cdef double dy2 = y_interp_ptr[1]
    cdef double dy3 = y_interp_ptr[2]
    cdef double e1  = y_interp_ptr[3]
    cdef double e2  = y_interp_ptr[4]
    cdef double e3  = y_interp_ptr[5]


    # Corrupt or otherwise change up the arg pointer
    args_ptr[0] = -99.0
    args_ptr[1] = -99.0
    args_ptr[2] = -99.0
    args_ptr = <double*>realloc(args_ptr, sizeof(double) * 50)
    cdef size_t i 
    for i in range(50):
        args_ptr[i] = -99.0
    

    # Recall the solution to see if it still produces the expected values
    result.get().call(check_t, y_interp_ptr)

    np.testing.assert_allclose(dy1, y_interp_ptr[0])
    np.testing.assert_allclose(dy2, y_interp_ptr[1])
    np.testing.assert_allclose(dy3, y_interp_ptr[2])
    np.testing.assert_allclose(e1, y_interp_ptr[3])
    np.testing.assert_allclose(e2, y_interp_ptr[4])
    np.testing.assert_allclose(e3, y_interp_ptr[5])

    free(args_ptr)

    return True

def cytester(
        int diffeq_number,
        tuple t_span = None,
        const double[::1] y0 = None,
        double[::1] args = None,
        int method = 1,
        size_t expected_size = 0,
        size_t max_num_steps = 0,
        size_t max_ram_MB = 2000,
        bint dense_output = False,
        double[::1] t_eval = None,
        double rtol = 1.0e-3, 
        double atol = 1.0e-6,
        double[::1] rtol_array = None,
        double[::1] atol_array = None,
        double max_step = MAX_STEP,
        double first_step = 0.0
        ):
    cdef size_t i
    cdef double* t_eval_ptr = NULL
    cdef size_t len_t_eval = 0
    if t_eval is not None:
        len_t_eval = len(t_eval)
        t_eval_ptr = &t_eval[0]

    cdef size_t num_y
    cdef double[10] y0_arr
    cdef double* y0_ptr = &y0_arr[0]

    cdef double[2] t_span_arr
    cdef double* t_span_ptr = &t_span_arr[0]

    cdef size_t num_extra = 0
    cdef DiffeqFuncType diffeq = NULL
    cdef PreEvalFunc pre_eval_func = NULL
    if diffeq_number == 0:
        diffeq = baseline_diffeq
    elif diffeq_number == 1:
        diffeq = accuracy_test_diffeq
    elif diffeq_number == 2:
        diffeq = extraoutput_test_diffeq
        num_extra = 2
    elif diffeq_number == 3:
        diffeq = lorenz_diffeq
    elif diffeq_number == 4:
        diffeq = lorenz_extraoutput_diffeq
        num_extra = 3
    elif diffeq_number == 5:
        diffeq = lotkavolterra_diffeq
    elif diffeq_number == 6:
        diffeq = pendulum_diffeq
    elif diffeq_number == 7:
        diffeq = arbitrary_arg_test
    elif diffeq_number == 8:
        diffeq = pendulum_preeval_diffeq
        pre_eval_func = pendulum_preeval_func
    else:
        raise NotImplementedError

    # Set up additional argument information
    cdef char* args_ptr = NULL
    cdef size_t size_of_args = 0

    # Arg double array for the diffeq's that use it
    cdef bint cast_arg_dbl   = False
    cdef double[10] args_arr
    cdef double* args_ptr_dbl = &args_arr[0]
    # Abitrary arg test requires a ArbitraryArgStruct class instance to be passed in
    cdef ArbitraryArgStruct arb_arg_struct = ArbitraryArgStruct(1.0, False, True, 1.0, 9.81)
    
    # Check if generic testing was requested.
    if y0 is None:
        # Generic Testing Requested
        
        # Use a higher tolerance for testing
        rtol = 1.0e-6
        atol = 1.0e-8

        # Get inputs for requested functions
        if diffeq_number == 0:
            num_y = 2
            y0_ptr[0] = 20.0
            y0_ptr[1] = 20.0
            t_span_ptr[0] = 0.0
            t_span_ptr[1] = 20.0

        elif diffeq_number == 1:
            num_y = 2
            y0_ptr[0] = 0.0
            y0_ptr[1] = 1.0
            t_span_ptr[0] = 0.0
            t_span_ptr[1] = 10.0

        elif diffeq_number == 2:
            num_y = 2
            y0_ptr[0] = 20.0
            y0_ptr[1] = 20.0
            t_span_ptr[0] = 0.0
            t_span_ptr[1] = 20.0

        elif diffeq_number == 3:
            num_y = 3
            y0_ptr[0] = 1.0
            y0_ptr[1] = 0.0
            y0_ptr[2] = 0.0
            t_span_ptr[0] = 0.0
            t_span_ptr[1] = 10.0
            args_ptr_dbl[0] = 10.0
            args_ptr_dbl[1] = 28.0
            args_ptr_dbl[2] = 8.0 / 3.0
            size_of_args = 8 * 3
            cast_arg_dbl = True
            
        elif diffeq_number == 4:
            num_y = 3
            y0_ptr[0] = 1.0
            y0_ptr[1] = 0.0
            y0_ptr[2] = 0.0
            t_span_ptr[0] = 0.0
            t_span_ptr[1] = 10.0
            args_ptr_dbl[0] = 10.0
            args_ptr_dbl[1] = 28.0
            args_ptr_dbl[2] = 8.0 / 3.0
            size_of_args = 8 * 3
            cast_arg_dbl = True

        elif diffeq_number == 5:
            num_y = 2
            y0_ptr[0] = 10.0
            y0_ptr[1] = 5.0
            t_span_ptr[0] = 0.0
            t_span_ptr[1] = 15.0
            args_ptr_dbl[0] = 1.5
            args_ptr_dbl[1] = 1.0
            args_ptr_dbl[2] = 3.0
            args_ptr_dbl[3] = 1.0
            size_of_args = 8 * 4
            cast_arg_dbl = True

        elif diffeq_number == 6:
            num_y = 2
            y0_ptr[0] = 0.01
            y0_ptr[1] = 0.0
            t_span_ptr[0] = 0.0
            t_span_ptr[1] = 10.0
            args_ptr_dbl[0] = 1.0
            args_ptr_dbl[1] = 1.0
            args_ptr_dbl[2] = 9.81
            size_of_args = 8 * 3
            cast_arg_dbl = True

        elif diffeq_number == 7:
            num_y = 2
            y0_ptr[0] = 0.01
            y0_ptr[1] = 0.0
            t_span_ptr[0] = 0.0
            t_span_ptr[1] = 10.0
            # Set args pointer to our arb args struct variable's address and cast it to a void pointer
            args_ptr = <char*>&arb_arg_struct
            size_of_args = sizeof(ArbitraryArgStruct)
            cast_arg_dbl = False
        
        elif diffeq_number == 8:
            num_y = 2
            y0_ptr[0] = 0.01
            y0_ptr[1] = 0.0
            t_span_ptr[0] = 0.0
            t_span_ptr[1] = 10.0
            args_ptr_dbl[0] = 1.0
            args_ptr_dbl[1] = 1.0
            args_ptr_dbl[2] = 9.81
            size_of_args = 8 * 3
            cast_arg_dbl = True

        else:
            raise NotImplementedError
    else:
        # Regular testing requested.
        num_y = len(y0)
        for i in range(num_y):
            y0_ptr[i] = y0[i]
        t_span_ptr[0] = t_span[0]
        t_span_ptr[1] = t_span[1]
        if args is not None:
            args_ptr     = <char*>&args[0]
            size_of_args = sizeof(double) * args.size
        else:
            args_ptr     = NULL
            size_of_args = 0

    if cast_arg_dbl:
        args_ptr     = <char*>args_ptr_dbl
        if size_of_args == 0:
            size_of_args = sizeof(args_arr)

    # Parse rtol
    cdef double* rtols_ptr = NULL
    if rtol_array is not None:
        rtols_ptr = &rtol_array[0]
    
    # Parse atol
    cdef double* atols_ptr = NULL
    if atol_array is not None:
        atols_ptr = &atol_array[0]

    cdef CySolveOutput result = cysolve_ivp(
        diffeq,
        t_span_ptr,
        y0_ptr,
        num_y,
        method = method,
        rtol = rtol,
        atol = atol,
        args_ptr = args_ptr,
        size_of_args = size_of_args,
        num_extra = num_extra,
        max_num_steps = max_num_steps,
        max_ram_MB = max_ram_MB,
        dense_output = dense_output,
        t_eval = t_eval_ptr,
        len_t_eval = len_t_eval,
        pre_eval_func = pre_eval_func,
        rtols_ptr = rtols_ptr,
        atols_ptr = atols_ptr,
        max_step = max_step,
        first_step = first_step,
        expected_size = expected_size
        )

    cdef WrapCySolverResult pysafe_result = WrapCySolverResult()
    pysafe_result.set_cyresult_pointer(result)

    return pysafe_result
